# -*- coding: utf-8 -*-

import os
import signal
import subprocess
import sys
import time
import webbrowser

import boto3
import click

from dagster import DagsterInvariantViolationError
from dagster.utils import dagster_home_dir

from .aws_util import (
    create_ec2_instance,
    create_key_pair,
    get_or_create_security_group,
    get_validated_ami_id,
    select_region,
    select_vpc,
)
from .config import HostConfig
from .term import run_remote_cmd, Term


# Client code will be deposited here on the remote EC2 instance
SERVER_CLIENT_CODE_HOME = '/opt/dagster/app/'


def get_dagster_home():
    '''Ensures that the user has set a valid DAGSTER_HOME in environment and that it exists
    '''
    try:
        dagster_home = dagster_home_dir()
    except DagsterInvariantViolationError:
        Term.fatal(
            '''DAGSTER_HOME is not set! Before continuing, set with e.g.:

export DAGSTER_HOME=~/.dagster

You may want to add this line to your .bashrc or .zshrc file.
'''
        )

    Term.info('Found DAGSTER_HOME in environment at: %s' % dagster_home)

    if not os.path.isdir(dagster_home):
        Term.fatal('The specified DAGSTER_HOME folder does not exist! Create before continuing.')
    return dagster_home


def exit_gracefully(_signum, _frame):
    '''Prevent stack trace spew on Ctrl-C
    '''
    click.echo(click.style('\n\nCommand killed by keyboard interrupt, quitting\n\n', fg='yellow'))
    sys.exit(1)


@click.group()
def main():
    signal.signal(signal.SIGINT, exit_gracefully)


@main.command()
def init():
    '''🚀 Initialize an EC2 VM to host Dagit'''
    click.echo('\n🌈 Welcome to Dagit + AWS quickstart cloud init!\n')

    # this ensures DAGSTER_HOME exists before we continue
    dagster_home = get_dagster_home()

    already_run = HostConfig.exists(dagster_home)
    prev_config = None
    if already_run:
        click.confirm(
            'dagster-aws has already been initialized! Continue?', default=False, abort=True
        )
        prev_config = HostConfig.load(dagster_home)

    # Get region
    if prev_config and prev_config.region:
        Term.success('Found existing region, continuing with %s' % prev_config.region)
        region = prev_config.region
    else:
        region = select_region()

    client = boto3.client('ec2', region_name=region)
    ec2 = boto3.resource('ec2', region_name=region)

    vpc = select_vpc(client, ec2)

    security_group_id = get_or_create_security_group(client, ec2, vpc)

    ami_id = get_validated_ami_id(client)

    if prev_config and os.path.exists(prev_config.key_file_path):
        Term.success(
            'Found existing key pair %s, continuing with %s'
            % (prev_config.key_pair_name, prev_config.key_file_path)
        )
        key_pair_name, key_file_path = prev_config.key_pair_name, prev_config.key_file_path
    else:
        key_pair_name, key_file_path = create_key_pair(client, dagster_home)

    inst = create_ec2_instance(ec2, security_group_id, ami_id, key_pair_name)

    # Save host configuration for future commands
    cfg = HostConfig(
        inst.public_dns_name,
        inst.id,
        region,
        security_group_id,
        key_pair_name,
        key_file_path,
        ami_id,
    )
    cfg.save(dagster_home)

    click.echo(
        click.style(
            '''🚀 To sync your Dagster project, in your project directory, run:

    dagster-aws up

You can also open a shell on your dagster-aws instance with:

    dagster-aws shell

For full details, see dagster-aws --help
            ''',
            fg='green',
        )
    )


@main.command()
def shell():
    '''Open an SSH shell on the remote server'''
    dagster_home = get_dagster_home()
    cfg = HostConfig.load(dagster_home)

    # Lands us directly in SERVER_CLIENT_CODE_HOME
    run_remote_cmd(cfg.key_file_path, cfg.public_dns_name, 'cd %s; bash' % SERVER_CLIENT_CODE_HOME)


@main.command()
def up():
    '''🌱 Sync your Dagster project to the remote server'''
    dagster_home = get_dagster_home()
    cfg = HostConfig.load(dagster_home)

    if cfg.local_path is None:
        cwd = os.getcwd()
        Term.info('Local path not configured; setting to %s' % cwd)
        cfg = cfg._replace(local_path=cwd)
        cfg.save(dagster_home)

    if not os.path.exists(os.path.join(cfg.local_path, 'repository.yaml')):
        Term.fatal('No repository.yaml found in %s, create before continuing.' % cfg.local_path)

    rsync_command = [
        'rsync',
        '-avL',
        '--progress',
        # Exclude a few common paths
        '--exclude',
        '\'.pytest_cache\'',
        '--exclude',
        '\'.git\'',
        '--exclude',
        '\'__pycache__\'',
        '--exclude',
        '\'*.pyc\'',
        '-e',
        '"ssh -i %s"' % cfg.key_file_path,
        '%s/*' % cfg.local_path,
        'ubuntu@%s:%s' % (cfg.public_dns_name, SERVER_CLIENT_CODE_HOME),
    ]
    Term.info('rsyncing local path %s to %s' % (cfg.local_path, cfg.public_dns_name))
    click.echo('\n' + ' '.join(rsync_command) + '\n')
    subprocess.call(' '.join(rsync_command), shell=True)

    Term.waiting('Testing that pipeline loads correctly on remote host...')
    retval = run_remote_cmd(
        cfg.key_file_path,
        cfg.public_dns_name,
        'export PYTHONPATH=$PYTHONPATH:/opt/dagster/app && '
        'source /opt/dagster/venv/bin/activate && '
        'cd /opt/dagster/app/ && '
        '/opt/dagster/venv/bin/dagster pipeline list',
    )

    if retval == 0:
        Term.success('Pipeline test succeeded')
    else:
        Term.fatal('Errors in pipeline; fix before proceeding')

    Term.waiting('Restarting dagit systemd service...')
    retval = run_remote_cmd(cfg.key_file_path, cfg.public_dns_name, 'sudo systemctl restart dagit')
    Term.success('Synchronization succeeded. Opening in browser...')

    # Open dagit in browser, but sleep for a few seconds first to give the service time to finish
    # restarting
    time.sleep(2)
    webbrowser.open_new_tab('http://%s:3000' % cfg.public_dns_name)


@main.command()
def update_dagster():
    '''Update the remote copy of Dagster'''
    dagster_home = get_dagster_home()
    cfg = HostConfig.load(dagster_home)

    Term.waiting(
        'Running a git pull and make rebuild_dagit on the remote dagster, this may take a while...'
    )
    retval = run_remote_cmd(
        cfg.key_file_path, cfg.public_dns_name, 'cd /opt/dagster/dagster && git pull'
    )
    if retval == 0:
        Term.info('Dagster git refresh completed! Rebuilding dagit...')
    else:
        Term.fatal('git pull failed')

    retval = run_remote_cmd(
        cfg.key_file_path,
        cfg.public_dns_name,
        'source /opt/dagster/venv/bin/activate && '
        'cd /opt/dagster/dagster/ && '
        'make rebuild_dagit',
    )
    if retval == 0:
        Term.success('Updating complete!')
    else:
        Term.fatal('Rebuilding dagit failed')


@main.command()
def nuke():
    '''💥 Terminate your EC2 instance'''
    dagster_home = get_dagster_home()

    already_run = HostConfig.exists(dagster_home)

    if not already_run:
        Term.fatal('No existing configuration detected, exiting')

    cfg = HostConfig.load(dagster_home)

    client = boto3.client('ec2', region_name=cfg.region)
    ec2 = boto3.resource('ec2', region_name=cfg.region)

    instances = ec2.instances.filter(InstanceIds=[cfg.instance_id])

    Term.warning('This will terminate the following: ')
    for instance in instances:
        name = None
        for tag in instance.tags:
            if tag.get('Key') == 'Name':
                name = tag.get('Value')
        click.echo('\t%s %s %s' % (instance.id, name, instance.instance_type))

    click.confirm('\nThis step cannot be undone. Continue?', default=False, abort=True)

    Term.waiting('Terminating...')
    instances.terminate()

    # Wipe all instance-related configs
    cfg = cfg._replace(public_dns_name=None, instance_id=None, ami_id=None, local_path=None)

    # Prompt user to remove key pair
    should_delete_key_pair = click.confirm(
        'Do you also want to remove the key pair %s?' % cfg.key_pair_name
    )
    if should_delete_key_pair:
        client.delete_key_pair(KeyName=cfg.key_pair_name)
        cfg = cfg._replace(key_pair_name=None, key_file_path=None)

    # Prompt user to delete security group also
    should_remove_security_group = click.confirm(
        'Do you also want to remove the security group %s?' % cfg.security_group_id
    )
    if should_remove_security_group:
        client.delete_security_group(GroupId=cfg.security_group_id)
        cfg = cfg._replace(security_group_id=None)

    # Write out updated config
    cfg.save(dagster_home)

    Term.success('Done!')



import datetime
import getpass
import os

import boto3
import click
import six

from .term import Spinner, Term

# Default to northern CA
DEFAULT_REGION = 'us-west-1'

# User can select an EC2 instance type if this is too small
DEFAULT_INSTANCE_TYPE = 't3.medium'

# We'll create this security group if it doesn't exist
DEFAULT_SECURITY_GROUP = 'dagit-sg'

# Ubuntu LTS
DEFAULT_AMI = 'ami-09eb5e8a83c7aa890'


def get_all_regions():
    '''Retrieves all regions/endpoints that work with EC2
    '''
    response = boto3.client('ec2').describe_regions()
    return sorted([r['RegionName'] for r in response['Regions']])


def select_region():
    '''User can select the region in which to instantiate EC2 machine.
    '''
    regions = click.Choice(get_all_regions())

    return click.prompt(
        'Select an AWS region ' + click.style('[default: %s]' % DEFAULT_REGION, fg='green'),
        type=regions,
        default=DEFAULT_REGION,
        show_default=False,
        show_choices=True,
    )


def select_vpc(client, ec2):
    '''User can select the VPC they would like to use for this machine.
    '''
    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    default_vpcs = list(ec2.vpcs.filter(Filters=filters))

    if not default_vpcs:
        Term.fatal('AWS account does not have a default VPC; needs manual setup')
    elif len(default_vpcs) > 1:
        Term.fatal('AWS account has multiple default VPCs; needs manual setup')

    default_vpc = default_vpcs[0]

    vpc_names = []
    for vpc in ec2.vpcs.iterator():
        response = client.describe_vpcs(VpcIds=[vpc.id])
        name = [tag['Value'] for tag in response['Vpcs'][0].get('Tags', []) if tag['Key'] == 'Name']
        vpc_names.append('%s (%s)' % (vpc.id, name[0]) if name else vpc.id)

    vpc = click.prompt(
        '\nSelect a VPC '
        + click.style('[default: %s]' % default_vpc.id, fg='green')
        + '\n\n'
        + click.style('  Existing VPCs:\n  ', fg='blue')
        + '\n  '.join(vpc_names),
        type=str,
        default=default_vpc,
        show_default=False,
    )
    return vpc


def get_or_create_security_group(client, ec2, vpc):
    '''Allows the user to select a security group, otherwise we create one for the dagit host.
    '''
    groups = [
        g for g in client.describe_security_groups()['SecurityGroups'] if g['VpcId'] == vpc.id
    ]

    for group in groups:
        if group['GroupName'] == DEFAULT_SECURITY_GROUP:
            Term.success(
                'Found existing dagit security group, continuing with %s' % group['GroupId']
            )
            return group['GroupId']

    groups_str = ['%s (New security group)' % DEFAULT_SECURITY_GROUP] + [
        '%s (%s)' % (g['GroupId'], g['GroupName']) for g in groups
    ]

    sg = click.prompt(
        '\nSelect a security group ID: '
        + click.style(
            '[default: create new security group %s]' % DEFAULT_SECURITY_GROUP, fg='green'
        )
        + '\n\n'
        + click.style('  Existing security groups:\n  ', fg='blue')
        + '\n  '.join(groups_str),
        type=str,
        default=DEFAULT_SECURITY_GROUP,
        show_default=False,
    )

    if sg == DEFAULT_SECURITY_GROUP:
        Term.waiting('Creating dagit security group...')
        group = ec2.create_security_group(
            GroupName=DEFAULT_SECURITY_GROUP, Description='dagit Security Group', VpcId=vpc.id
        )
        group.authorize_ingress(
            GroupName=DEFAULT_SECURITY_GROUP,
            IpPermissions=[
                {
                    'IpProtocol': 'tcp',
                    'FromPort': 3000,
                    'ToPort': 3000,
                    'IpRanges': [{'CidrIp': '0.0.0.0/0', 'Description': 'HTTP'}],
                },
                {
                    'IpProtocol': 'tcp',
                    'FromPort': 22,
                    'ToPort': 22,
                    'IpRanges': [{'CidrIp': '0.0.0.0/0', 'Description': 'SSH'}],
                },
            ],
        )
        group.create_tags(Tags=[{'Key': 'Name', 'Value': DEFAULT_SECURITY_GROUP}])
        Term.success('Security group created!')

    return sg


def create_key_pair(client, dagster_home):
    '''Ensures we have a key pair in place for creating EC2 instance so that the user can SSH to the
    machine once it is created.
    '''
    now = datetime.datetime.utcnow()
    default_key_pair_name = 'dagster-keypair-%s-%s' % (
        getpass.getuser(),
        now.strftime("%Y%m%dT%H%M%S"),
    )

    Term.waiting('Creating new key pair...')

    existing_key_pairs = [k['KeyName'] for k in client.describe_key_pairs()['KeyPairs']]

    key_pair_name = None
    while not key_pair_name:
        key_pair_name = click.prompt(
            'Key pair name ' + click.style('[default is %s]' % default_key_pair_name, fg='green'),
            type=str,
            default=default_key_pair_name,
            show_default=False,
        )

        # Ensure key doesn't already exist
        if key_pair_name in existing_key_pairs:
            Term.error('Specified key pair already exists, won\'t create')
            key_pair_name = None

        else:
            # key does not exist yet, safe to create
            keypair = client.create_key_pair(KeyName=key_pair_name)

            key_directory = os.path.join(dagster_home, 'keys')
            key_file_path = os.path.join(key_directory, '%s.pem' % key_pair_name)

            # Ensure key directory exists
            if not os.path.exists(key_directory):
                os.makedirs(key_directory)

            # Write key material and ensure appropriate permissions are set
            with open(key_file_path, 'wb') as f:
                f.write(six.ensure_binary(keypair['KeyMaterial']))
            os.chmod(key_file_path, 0o600)

            Term.success(
                'Key pair %s created and saved to local file %s!' % (key_pair_name, key_file_path)
            )
            return key_pair_name, key_file_path


def get_validated_ami_id(_client):
    '''Permit the user to select an AMI to use. This is currently unused, as instead we default to
    using a Ubuntu 16.04 LTS AMI.
    '''

    # For now, we will force users to use a specific Ubuntu AMI
    return DEFAULT_AMI

    # ami_id = None
    # while not ami_id:
    #     ami_id = click.prompt(
    #         '\nChoose an AMI to use (must be Debian-based) '
    #         + click.style('[default is %s]' % DEFAULT_AMI, fg='green'),
    #         type=str,
    #         default=DEFAULT_AMI,
    #         show_default=False,
    #     )

    #     # Boto will throw a ClientError exception if this AMI doesn't exist.
    #     try:
    #         client.describe_images(ImageIds=[ami_id])
    #     except ClientError:
    #         Term.error('Specified AMI does not exist, fix to continue')
    #         ami_id = None

    # return ami_id


def create_ec2_instance(ec2, security_group_id, ami_id, key_pair_name):
    '''Actually create the EC2 instance given the provided configuration.
    '''
    click.echo('\n🌱 Provisioning an EC2 instance for dagit')

    instance_type = click.prompt(
        '\nChoose an EC2 instance type '
        + click.style('[default is %s]' % DEFAULT_INSTANCE_TYPE, fg='green'),
        type=str,
        default=DEFAULT_INSTANCE_TYPE,
        show_default=False,
    )

    ec2_default_name = 'dagit-ec2-%s' % getpass.getuser()
    ec2_instance_name = click.prompt(
        '\nChoose an EC2 instance name '
        + click.style('[default is %s]' % ec2_default_name, fg='green'),
        type=str,
        default=ec2_default_name,
        show_default=False,
    )

    Term.waiting('Creating dagit EC2 instance...')
    click.echo(
        '  '
        + '\n  '.join(
            [
                'Name: %s' % ec2_instance_name,
                'AMI: %s' % ami_id,
                'Instance Type: %s' % instance_type,
                'Security Group: %s' % security_group_id,
                'Key Pair: %s' % key_pair_name,
            ]
        )
    )

    # Here we actually create the EC2 instance
    with Spinner():
        with open(os.path.join(os.path.dirname(__file__), 'shell', 'init.sh'), 'rb') as f:
            init_script = six.ensure_str(f.read())

        instances = ec2.create_instances(
            ImageId=ami_id,
            InstanceType=instance_type,
            MaxCount=1,
            MinCount=1,
            SecurityGroupIds=[security_group_id],
            KeyName=key_pair_name,
            UserData=init_script,
        )
        inst = instances[0]
        inst.wait_until_running()

        # Add name tag to instance after creation
        ec2.create_tags(Resources=[inst.id], Tags=[{'Key': 'Name', 'Value': ec2_instance_name}])

        # Load details like public DNS name, etc.
        inst.reload()

    Term.success('dagit EC2 instance %s launched!' % inst.id)
    return inst



import os
import yaml

from dagster import seven

from dagster_aws.cli.config import HOST_CONFIG_FILE, HostConfig


def test_host_config():
    cfg = HostConfig(
        public_dns_name='foo',
        region='us-west-1',
        security_group_id='sg-12345',
        key_pair_name='foobar',
        key_file_path='/some/path',
        ami_id='ami-12345',
    )

    tmp_dir = seven.get_system_temp_directory()
    outfile = os.path.join(tmp_dir, HOST_CONFIG_FILE)

    cfg.save(tmp_dir)

    with open(outfile) as f:
        parsed = yaml.load(f)

    assert 'dagit-aws-host' in parsed
    configs = parsed['dagit-aws-host']
    assert configs['public_dns_name'] == 'foo'
    assert configs['region'] == 'us-west-1'
    assert configs['security_group_id'] == 'sg-12345'
    assert configs['key_pair_name'] == 'foobar'
    assert configs['key_file_path'] == '/some/path'
    assert configs['ami_id'] == 'ami-12345'

    res = cfg.load(tmp_dir)
    assert res == HostConfig(**configs)



import os
import yaml

from collections import namedtuple

import six

from dagster import check

from .term import Term


HOST_CONFIG_FILE = '.dagit-aws-config'


class HostConfig(
    namedtuple(
        '_HostConfig',
        'public_dns_name instance_id region security_group_id key_pair_name key_file_path ami_id local_path',
    )
):
    '''Serialize the user's AWS host configuration to a YAML file for future use.
    '''

    def __new__(
        cls,
        public_dns_name=None,
        instance_id=None,
        region=None,
        security_group_id=None,
        key_pair_name=None,
        key_file_path=None,
        ami_id=None,
        local_path=None,
    ):
        return super(HostConfig, cls).__new__(
            cls,
            public_dns_name=check.opt_str_param(public_dns_name, 'public_dns_name'),
            instance_id=check.opt_str_param(instance_id, 'instance_id'),
            region=check.opt_str_param(region, 'region'),
            security_group_id=check.opt_str_param(security_group_id, 'security_group_id'),
            key_pair_name=check.opt_str_param(key_pair_name, 'key_pair_name'),
            key_file_path=check.opt_str_param(key_file_path, 'key_file_path'),
            ami_id=check.opt_str_param(ami_id, 'ami_id'),
            local_path=check.opt_str_param(local_path, 'local_path'),
        )

    @staticmethod
    def exists(dagster_home):
        cfg_path = os.path.join(dagster_home, HOST_CONFIG_FILE)
        return os.path.exists(cfg_path)

    def save(self, dagster_home):
        # Save configuration to a file for future use
        cfg_path = os.path.join(dagster_home, HOST_CONFIG_FILE)
        with open(cfg_path, 'wb') as f:
            output_record = {
                'dagit-aws-host': {
                    'public_dns_name': self.public_dns_name,
                    'instance_id': self.instance_id,
                    'region': self.region,
                    'security_group_id': self.security_group_id,
                    'key_pair_name': self.key_pair_name,
                    'key_file_path': self.key_file_path,
                    'ami_id': self.ami_id,
                    'local_path': self.local_path,
                }
            }
            f.write(six.ensure_binary(yaml.dump(output_record, default_flow_style=False)))
        Term.info('Saved host configuration to %s' % cfg_path)

    @staticmethod
    def load(dagster_home):
        cfg_path = os.path.join(dagster_home, HOST_CONFIG_FILE)
        with open(cfg_path, 'rb') as f:
            raw_cfg = yaml.load(f)
        return HostConfig(**raw_cfg['dagit-aws-host'])
